#include "mockup.hpp"
#include <clocale>

#define TRACE puts

CDirChangeHandler::CDirChangeHandler()
    : m_cRefs(1)
    , m_pWatcher(NULL)
{
}

BOOL CDirChangeHandler::UnwatchDir()
{
    BOOL ret = TRUE;
    Lock();
    if (m_pWatcher)
        ret = m_pWatcher->UnwatchDir(this);
    Unlock();
    return ret;
}

CDirChangeHandler::~CDirChangeHandler()
{
    UnwatchDir();
    ::DeleteCriticalSection(&m_csLock);
}

ULONG CDirChangeHandler::AddRef()
{
    return ::InterlockedIncrement(&m_cRefs);
}

ULONG CDirChangeHandler::Release()
{
    ULONG cRefs = ::InterlockedDecrement(&m_cRefs);
    if (cRefs == 0)
    {
        delete this;
        return 0;
    }
    return cRefs;
}

BOOL CDirChangeHandler::OnChangeNotify(DWORD dwAction, LPCWSTR file1, LPCWSTR file2)
{
    switch (dwAction)
    {
    case FILE_ACTION_ADDED:
        printf("FILE_ACTION_ADDED: %ls, %ls\n", file1, file2);
        break;
    case FILE_ACTION_REMOVED:
        printf("FILE_ACTION_REMOVED: %ls, %ls\n", file1, file2);
        break;
    case FILE_ACTION_MODIFIED:
        //printf("FILE_ACTION_MODIFIED: %ls, %ls\n", file1, file2);
        break;
    case FILE_ACTION_RENAMED_NEW_NAME:
        printf("FILE_ACTION_RENAMED_NEW_NAME: %ls, %ls\n", file1, file2);
        break;
    default:
        printf("%08X: %ls, %ls\n", dwAction, file1, file2);
        break;
    }
    fflush(stdout);
    return TRUE;
}

void CDirChangeHandler::SetWatcher(CDirChangeWatcher *pWatcher)
{
    Lock();

    if (m_pWatcher && m_pWatcher != pWatcher)
        UnwatchDir();

    m_pWatcher = pWatcher;

    Unlock();
}

CDirWatchItem::CDirWatchItem(LPCWSTR pszDir, HANDLE hDir, CDirChangeHandler *pChangeHandler,
                             DWORD dwChangeFilter, BOOL bWatchSubDir)
    : m_strDir(pszDir)
    , m_hDir(hDir)
    , m_pChangeHandler(pChangeHandler)
    , m_dwChangeFilter(dwChangeFilter)
    , m_bWatchSubDir(bWatchSubDir)
    , m_hStartStopEvent(::CreateEvent(NULL, TRUE, FALSE, NULL))
    , m_RunningState(RUNNING_STATE_NOT_SET)
    , m_dwReadDirError(ERROR_SUCCESS)
{
    TRACE("CDirWatchItem");
    m_abBuffer[0] = 0;
    assert(m_hDir != INVALID_HANDLE_VALUE);
    assert(m_hStartStopEvent != NULL);
    m_pChangeHandler->AddRef();
}

CDirWatchItem::~CDirWatchItem()
{
    if (m_pChangeHandler)
        m_pChangeHandler->Release();
}

void CDirWatchItem::DeleteSelf(CDirChangeWatcher *pWatcher)
{
    TRACE("DeleteSelf");
    if (pWatcher)
        pWatcher->UnwatchDir(m_pChangeHandler);
    delete this;
}

DWORD CDirWatchItem::StartMonitor(HANDLE hCompPort)
{
    TRACE("StartMonitor");
    assert(hCompPort);

    Lock();
    m_RunningState = RUNNING_STATE_START_MONITORING;
    ::PostQueuedCompletionStatus(hCompPort,
                                 sizeof(this), reinterpret_cast<ULONG_PTR>(this),
                                 &m_Overlapped);
    Unlock();

    DWORD dwWait = 0;
    do
    {
        dwWait = ::WaitForSingleObject(m_hStartStopEvent, 500);
        TRACE("waiting");
    } while (dwWait != WAIT_OBJECT_0);

    ::ResetEvent(m_hStartStopEvent);
    return m_dwReadDirError;
}

BOOL CDirWatchItem::StopMonitor(HANDLE hCompPort)
{
    TRACE("StopMonitor");
    assert(hCompPort);
    if (SignalShutdown(hCompPort))
        return WaitForShutdown();
    return FALSE;
}

BOOL CDirWatchItem::SignalShutdown(HANDLE hCompPort)
{
    TRACE("SignalShutdown");
    BOOL ret = FALSE;
    assert(hCompPort);
    assert(m_hDir != INVALID_HANDLE_VALUE);

    Lock();
    m_RunningState = RUNNING_STATE_STOP;
    ret = ::PostQueuedCompletionStatus(hCompPort, sizeof(this),
                                       reinterpret_cast<ULONG_PTR>(this),
                                       &m_Overlapped);
    Unlock();

    return ret;
}

BOOL CDirWatchItem::WaitForShutdown()
{
    TRACE("WaitForShutdown");
    assert(m_hStartStopEvent);

    DWORD dwWait;
    MSG msg;
    do
    {
        TRACE("waiting 2");

        dwWait = ::MsgWaitForMultipleObjects(1, &m_hStartStopEvent, FALSE, 1000, QS_ALLINPUT);
        switch (dwWait)
        {
        case WAIT_OBJECT_0:
            break;

        case WAIT_OBJECT_0 + 1:
            while (::PeekMessageW(&msg, NULL, 0, 0, PM_REMOVE))
            {
                if (msg.message == WM_QUIT)
                    break;
                TranslateMessage(&msg);
                DispatchMessageW(&msg);
            }
            break;

        case WAIT_TIMEOUT:
            break;
        }
    } while (dwWait != WAIT_OBJECT_0);

    ::ResetEvent(m_hStartStopEvent);

    return dwWait == WAIT_OBJECT_0;
}

void CDirWatchItem::CloseDirectoryHandle()
{
    TRACE("CloseDirectoryHandle");
    if (m_hDir != INVALID_HANDLE_VALUE)
    {
        CloseHandle(m_hDir);
        m_hDir = INVALID_HANDLE_VALUE;
    }
}

CDirChangeWatcher::CDirChangeWatcher()
    : m_hCompPort(NULL)
    , m_hThread(NULL)
{
}

CDirChangeWatcher::~CDirChangeWatcher()
{
    UnwatchAll();
    CloseCompPort();
    CloseThread();
}

void CDirChangeWatcher::CloseCompPort()
{
    if (m_hCompPort)
    {
        ::CloseHandle(m_hCompPort);
        m_hCompPort = NULL;
    }
}

DWORD CDirChangeWatcher::WatchProc()
{
    TRACE("WatchProc");

    assert(m_hCompPort);

    DWORD cb;
    CDirWatchItem *pWatchItem;
    LPOVERLAPPED pOverlapped;

    do
    {
        TRACE("waiting in WatchProc");

        if (!::GetQueuedCompletionStatus(m_hCompPort, &cb, (PULONG_PTR)&pWatchItem,
                                         &pOverlapped, INFINITE))
        {
            assert(0);
            continue;
        }

        assert(pWatchItem);

        if (pWatchItem)
        {
            pWatchItem->Lock();
            RUNNING_STATE state = pWatchItem->m_RunningState;
            pWatchItem->Unlock();

            switch (state)
            {
            case RUNNING_STATE_START_MONITORING:
                TRACE("RUNNING_STATE_START_MONITORING");
                if (::ReadDirectoryChangesW(pWatchItem->m_hDir,
                                            pWatchItem->m_abBuffer,
                                            sizeof(pWatchItem->m_abBuffer),
                                            pWatchItem->m_bWatchSubDir,
                                            pWatchItem->m_dwChangeFilter,
                                            NULL,
                                            &pWatchItem->m_Overlapped,
                                            NULL))
                {
                    pWatchItem->m_RunningState = RUNNING_STATE_NORMAL;
                    pWatchItem->m_dwReadDirError = ERROR_SUCCESS;
                }
                else
                {
                    pWatchItem->m_dwReadDirError = ::GetLastError();
                }
                ::SetEvent(pWatchItem->m_hStartStopEvent);
                break;

            case RUNNING_STATE_STOP:
                TRACE("RUNNING_STATE_STOP");
                pWatchItem->CloseDirectoryHandle();
                ::SetEvent(pWatchItem->m_hStartStopEvent);
                break;

            case RUNNING_STATE_NORMAL:
                TRACE("RUNNING_STATE_NORMAL");

                ProcessChangeNotifications(pWatchItem);

                if (::ReadDirectoryChangesW(pWatchItem->m_hDir,
                                            pWatchItem->m_abBuffer,
                                            sizeof(pWatchItem->m_abBuffer),
                                            pWatchItem->m_bWatchSubDir,
                                            pWatchItem->m_dwChangeFilter,
                                            NULL,
                                            &pWatchItem->m_Overlapped,
                                            NULL))
                {
                    pWatchItem->m_dwReadDirError = ERROR_SUCCESS;
                }
                else
                {
                    pWatchItem->m_dwReadDirError = GetLastError();

                    INT nOldPriority = ::GetThreadPriority(::GetCurrentThread());
                    ::SetThreadPriority(::GetCurrentThread(), THREAD_PRIORITY_HIGHEST);

                    UnwatchDir(pWatchItem);

                    ::SetThreadPriority(::GetCurrentThread(), nOldPriority);
                }
                break;
            }
        }
    } while (pWatchItem);

    return 0;
}

std::wstring DoGetPathName(CDirWatchItem *pWatchItem, FILE_NOTIFY_INFORMATION *pFNI)
{
    WCHAR szPath[MAX_PATH] = L"";
    SIZE_T cbMax;

    if (pFNI->FileNameLength >= MAX_PATH * sizeof(WCHAR))
        cbMax = MAX_PATH * sizeof(WCHAR);
    else
        cbMax = pFNI->FileNameLength;

    CopyMemory(szPath, pFNI->FileName, cbMax);
    szPath[_countof(szPath) - 1] = 0;

    std::wstring strPath = pWatchItem->m_strDir;
    if (strPath.size() && strPath[strPath.size() - 1] != L'\\')
        strPath += L'\\';
    strPath += szPath;

    return strPath;
}

void CDirChangeWatcher::ProcessChangeNotifications(CDirWatchItem *pWatchItem)
{
    TRACE("ProcessChangeNotifications");

    if (!pWatchItem || ::IsBadWritePtr(pWatchItem, sizeof(*pWatchItem)))
    {
        assert(0);
        return;
    }

    CDirChangeHandler *pChangeHandler = pWatchItem->m_pChangeHandler;
    if (!pChangeHandler)
    {
        assert(0);
        return;
    }

    LPBYTE pb = pWatchItem->m_abBuffer;
    size_t ib = 0;
    std::wstring strPath, strOldPath;

    for (;;)
    {
        FILE_NOTIFY_INFORMATION *pFNI = reinterpret_cast<FILE_NOTIFY_INFORMATION*>(pb);

        switch (pFNI->Action)
        {
        case FILE_ACTION_ADDED:
        case FILE_ACTION_REMOVED:
        case FILE_ACTION_MODIFIED:
        case FILE_ACTION_RENAMED_NEW_NAME:
            strPath = DoGetPathName(pWatchItem, pFNI);
            if (pFNI->Action == FILE_ACTION_RENAMED_NEW_NAME)
                pChangeHandler->OnChangeNotify(pFNI->Action, strOldPath.c_str(), strPath.c_str());
            else
                pChangeHandler->OnChangeNotify(pFNI->Action, strPath.c_str(), NULL);
            break;

        case FILE_ACTION_RENAMED_OLD_NAME:
            strOldPath = DoGetPathName(pWatchItem, pFNI);
            break;
        }

        if (pFNI->NextEntryOffset == 0)
            break;

        pb += pFNI->NextEntryOffset;
        ib += pFNI->NextEntryOffset;
        assert(ib <= sizeof(pWatchItem->m_abBuffer));
    }
}

void CDirChangeWatcher::CloseThread()
{
    if (m_hThread)
    {
        ::CloseHandle(m_hThread);
        m_hThread = NULL;
    }
}

BOOL CDirChangeWatcher::IsWatchingDir(LPCWSTR pszDir)
{
    Lock();
    INT iItem = FindWatchItem(pszDir);
    Unlock();
    return iItem >= 0;
}

DWORD
CDirChangeWatcher::WatchDir(LPCWSTR pszDir, DWORD dwChangeFilter,
                            CDirChangeHandler *pChangeHandler, BOOL bWatchSubDir)
{
    TRACE("WatchDir");

    if (!pszDir || !pszDir[0] || !pChangeHandler)
    {
        assert(0);
        return ERROR_INVALID_PARAMETER;
    }

    if (!::PathIsDirectoryW(pszDir))
    {
        assert(0);
        return ERROR_BAD_PATHNAME;
    }

    if (IsWatchingDir(pszDir))
        UnwatchDir(pszDir);

    HANDLE hDir = ::CreateFileW(pszDir,
                                FILE_LIST_DIRECTORY,
                                FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
                                NULL,
                                OPEN_EXISTING,
                                FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OVERLAPPED,
                                NULL);
    if (hDir == INVALID_HANDLE_VALUE)
    {
        assert(0);
        return GetLastError();
    }

    CDirWatchItem *pWatchItem = new CDirWatchItem(pszDir, hDir, pChangeHandler, dwChangeFilter, bWatchSubDir);

    m_hCompPort = ::CreateIoCompletionPort(hDir, m_hCompPort,
                                           reinterpret_cast<ULONG_PTR>(pWatchItem), 0);
    if (!m_hCompPort)
    {
        DWORD dwError = GetLastError();
        pWatchItem->DeleteSelf(NULL);
        assert(0);
        return dwError;
    }

    if (!m_hThread)
        m_hThread = ::CreateThread(NULL, 0, WatchThreadProc, this, 0, NULL);

    if (!m_hThread)
    {
        assert(0);
        return ERROR_MAX_THRDS_REACHED;
    }

    DWORD dwError = pWatchItem->StartMonitor(m_hCompPort);
    if (dwError != ERROR_SUCCESS)
    {
        assert(0);
        pWatchItem->DeleteSelf(NULL);
        return dwError;
    }

    AddWatchItem(pWatchItem);

    pChangeHandler->SetWatcher(this);

    TRACE("OK");
    return ERROR_SUCCESS;
}

INT CDirChangeWatcher::AddWatchItem(CDirWatchItem *pWatchItem)
{
    Lock();
    for (size_t iItem = 0; iItem < m_WatchItems.size(); ++iItem)
    {
        if (m_WatchItems[iItem] == NULL)
        {
            m_WatchItems[iItem] = pWatchItem;
            Unlock();
            return static_cast<INT>(iItem);
        }
    }
    size_t iItem = m_WatchItems.size();
    m_WatchItems.push_back(pWatchItem);
    Unlock();
    return iItem;
}

INT CDirChangeWatcher::FindWatchItem(LPCWSTR pszDir) const
{
    for (size_t iItem = 0; iItem < m_WatchItems.size(); ++iItem)
    {
        CDirWatchItem *pWatchItem = m_WatchItems[iItem];
        if (lstrcmpiW(pWatchItem->m_strDir.c_str(), pszDir) == 0)
            return static_cast<INT>(iItem);
    }
    return -1;
}

INT CDirChangeWatcher::FindWatchItem(CDirWatchItem *pWatchItem) const
{
    for (size_t iItem = 0; iItem < m_WatchItems.size(); ++iItem)
    {
        if (m_WatchItems[iItem] == pWatchItem)
            return static_cast<INT>(iItem);
    }
    return -1;
}

INT CDirChangeWatcher::FindWatchItem(CDirChangeHandler *pChangeHandler) const
{
    for (size_t iItem = 0; iItem < m_WatchItems.size(); ++iItem)
    {
        CDirWatchItem *pWatchItem = m_WatchItems[iItem];
        if (pWatchItem->m_pChangeHandler == pChangeHandler)
            return static_cast<INT>(iItem);
    }
    return -1;
}

BOOL CDirChangeWatcher::UnwatchDir(LPCWSTR pszDir)
{
    if (!m_hCompPort)
        return FALSE;

    BOOL ret = FALSE;

    Lock();
    INT iItem = FindWatchItem(pszDir);
    if (iItem >= 0)
    {
        CDirWatchItem *pWatchItem = m_WatchItems[iItem];
        m_WatchItems[iItem] = NULL;
        pWatchItem->StopMonitor(m_hCompPort);
        pWatchItem->DeleteSelf(this);
        ret = TRUE;
    }
    Unlock();

    return ret;
}

BOOL CDirChangeWatcher::UnwatchDir(CDirWatchItem *pWatchItem)
{
    BOOL ret = FALSE;

    Lock();
    INT iItem = FindWatchItem(pWatchItem);
    if (iItem >= 0)
    {
        m_WatchItems[iItem] = NULL;
        pWatchItem->StopMonitor(m_hCompPort);
        pWatchItem->DeleteSelf(this);
        ret = TRUE;
    }
    Unlock();

    return ret;
}

BOOL CDirChangeWatcher::UnwatchDir(CDirChangeHandler *pChangeHandler)
{
    BOOL ret = FALSE;

    Lock();
    INT iItem = FindWatchItem(pChangeHandler);
    if (iItem >= 0)
    {
        CDirWatchItem *pWatchItem = m_WatchItems[iItem];
        m_WatchItems[iItem] = NULL;
        pWatchItem->StopMonitor(m_hCompPort);
        pWatchItem->DeleteSelf(this);
        ret = TRUE;
    }
    Unlock();

    return ret;
}

BOOL CDirChangeWatcher::UnwatchAll()
{
    TRACE("UnwatchAll");

    if (m_hThread)
    {
        assert(m_hCompPort);
        Lock();

        for (size_t iItem = 0; iItem < m_WatchItems.size(); ++iItem)
        {
            CDirWatchItem *pWatchItem = m_WatchItems[iItem];
            m_WatchItems[iItem] = NULL;
            pWatchItem->StopMonitor(m_hCompPort);
            pWatchItem->DeleteSelf(this);
        }
        m_WatchItems.clear();

        ::PostQueuedCompletionStatus(m_hCompPort, 0, 0, NULL);

        ::WaitForSingleObject(m_hThread, INFINITE);
        m_hThread = NULL;

        CloseCompPort();

        Unlock();
        return TRUE;
    }

    return FALSE;
}

int main(void)
{
    setlocale(LC_ALL, "");

    CDirChangeWatcher watcher;
    CDirChangeHandler *pHandler = new CDirChangeHandler();
    BOOL bWatchSubDir = TRUE;

    LPCWSTR pszDir = L"mydir";
    DWORD dwChangeFilter =
        FILE_NOTIFY_CHANGE_FILE_NAME |
        FILE_NOTIFY_CHANGE_DIR_NAME |
        FILE_NOTIFY_CHANGE_ATTRIBUTES |
        FILE_NOTIFY_CHANGE_SIZE |
        FILE_NOTIFY_CHANGE_LAST_WRITE |
        FILE_NOTIFY_CHANGE_LAST_ACCESS |
        FILE_NOTIFY_CHANGE_CREATION |
        FILE_NOTIFY_CHANGE_SECURITY;

    if (watcher.IsWatchingDir(pszDir))
        watcher.UnwatchDir(pszDir);

    watcher.WatchDir(pszDir, dwChangeFilter, pHandler, bWatchSubDir);

    for (;;)
    {
        fflush(stdout);
        Sleep(1000);
    }

    return 0;
}
